# noqa: E501

from types import SimpleNamespace as obj
from tempfile import TemporaryDirectory, NamedTemporaryFile
from .. import check, checkSync, NotAGitRepoException
from shutil import rmtree
from os import path
import asyncio
import subprocess


x = "\x1b[31m✘\x1b[0m"
o = "\x1b[32m✔\x1b[0m"


def runTests():
    return asyncio.run(_runTests())


async def _runTests():
    # sync tests
    with TemporaryDirectory() as tmpdirname:
        for step in makeSteps(tmpdirname):
            if not (await step()):
                return False

    return True


# ------- #
# Helpers #
# ------- #


def makeSteps(cwd):
    testIsClean = makeTestIsClean(cwd)
    runSync = makeRunSync(cwd)
    runGitInit = lambda: runSync(["git", "init"])
    runGitAddAll = lambda: runSync(["git", "add", "."])

    def runResetGitDir():
        rmtree(path.join(cwd, ".git"))
        runGitInit()

    async def beforeGitInit():
        isClean = makeIsClean(cwd)
        try:
            await isClean.nonsync()
            print(
                f"  {x} check({cwd}) did not raise an exception before git init"
            )
            return False
        except NotAGitRepoException:
            pass
        except:
            print(
                f"  {x} check({cwd}) raised an incorrect exception before"
                " git init"
            )
            return False

        try:
            isClean.sync()
            print(
                f"  {x} checkSync({cwd}) did not raise an exception before"
                " git init"
            )
            return False
        except NotAGitRepoException:
            pass
        except:
            print(
                f"  {x} checkSync({cwd}) raised an incorrect exception before"
                " git init"
            )
            return False

        print(f"  {o} before git init")
        return True

    async def gitInit():
        runGitInit()
        return await testIsClean(True, "a newly initialized repo")

    async def unstaged():
        with NamedTemporaryFile(dir=cwd):
            return await testIsClean(False, "unstaged changes")

    async def staged():
        with NamedTemporaryFile(dir=cwd):
            runGitAddAll()
            result = await testIsClean(False, "staged changes")

        runResetGitDir()
        return result

    async def committed():
        with NamedTemporaryFile(dir=cwd):
            runGitAddAll()
            runSync(["git", "commit", "-m", '"initial commit"'])
            result = await testIsClean(True, "committed changes")

        runResetGitDir()
        return result

    return [beforeGitInit, gitInit, unstaged, staged, committed]


def makeTestIsClean(cwd):
    isClean = makeIsClean(cwd)

    async def testIsClean(expected, context):
        method = "check"
        actual = await isClean.nonsync()
        if expected == actual:
            method = "checkSync"
            actual = isClean.sync()

        if expected != actual:
            print(f"  {x} {method}({cwd}) returned {actual} with {context}")
        else:
            print(f"  {o} {context}")

        return expected == actual

    return testIsClean


def makeIsClean(cwd):
    return obj(sync=lambda: checkSync(cwd), nonsync=lambda: check(cwd))


def makeRunSync(cwd):
    return lambda cmd: subprocess.run(
        cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd
    )



from asyncio import gather
from types import SimpleNamespace as o


#
# I like to think the common use-case for 'gather' is to cancel all async
#   operations should any one of them fail.  This is not how 'gather' was
#   implemented though which is why I wrote this function
#
async def resolveAll(awaitables):
    gatherAll = None
    try:
        gatherAll = gather(*awaitables)
        result = await gatherAll
    finally:
        if gatherAll:
            gatherAll.cancel()

    return result


#
# I have no idea what to name this.  Python just doesn't have a clean ternary
#   operator which stinks and I didn't want to use 'iif' because that's
#   unreadable.  So for now we have a verbose and less performant, but readable
#   function chain
#


def whenTruthy(condition):
    def return_(truthyResult):
        def otherwise(falseyResult):
            if condition:
                return truthyResult
            else:
                return falseyResult

        return o(otherwise=otherwise)

    return o(return_=return_)


def iif(condition, whenTruthy, whenFalsey):
    if condition:
        return whenTruthy
    else:
        return whenFalsey



# ------- #
# Imports #
# ------- #

import os
from traceback import format_exc
from textwrap import dedent
from types import SimpleNamespace as o
from .check import checkSync, NotAGitRepoException
from .utils import iif


# ---- #
# Init #
# ---- #

options = set(["--dir", "--silent"])

twoLineSeps = os.linesep + os.linesep

usage = dedent(
    f"""
    Usage
    is-git-repo-clean [--dir <path>] [--silent]
    is-git-repo-clean --help

    Options
      dir:      path to the git repo to test.  Defaults to `os.getcwd()`
      silent:   disables output
      help:     print this
    """
)


# ---- #
# Main #
# ---- #


def getIsGitRepoClean(args):
    result = o(stdout=None, stderr=None, code=None)

    numArgs = len(args)
    if numArgs == 1 and args[0] == "--help":
        result.stdout = usage
        result.code = 0
        return result

    argsObj = validateAndParseArgs(args, result)

    if argsObj is result:
        return result

    isSilent = argsObj.silent

    try:
        isClean = checkSync(argsObj.dir or None)
        if not isSilent:
            if isClean:
                result.stdout = "yes"
            else:
                result.stderr = "no"

        result.code = iif(isClean, 0, 1)
        return result

    except NotAGitRepoException:
        if not isSilent:
            result.stderr = "dir is not a git repository"

        result.code = 3
        return result

    except Exception:
        if not isSilent:
            result.stderr = (
                f"unexpected error occurred{twoLineSeps}" + format_exc()
            )

        result.code = 4
        return result


# ------- #
# Helpers #
# ------- #


def validateAndParseArgs(args, result):
    i = 0
    argsObj = o(dir=None, silent=None)

    while i < len(args):
        arg = args[i]
        if arg not in options:
            if arg == "--help":
                result.stderr = (
                    "'--help' must be the only argument when passed"
                )
            else:
                result.stderr = f"invalid option '{arg}'"

            result.stderr += os.linesep + usage
            result.code = 2

            return result

        if arg == "--dir":
            if i == len(args) - 1:
                result.stderr = "'--dir' must be given a value"
                result.stderr += os.linesep + usage
                result.code = 2

                return result

            argsObj.dir = args[i + 1]
            i += 1
        else:
            argsObj.silent = True

        i += 1

    return argsObj
